/*
 * Quick & Dirty Config Utility for: CRWU-PC/IP
 *
 * This program operates on TCPIP.EXE - in general, you should:
 *     > COPY TCPIPSTA.EXE TCPIP.EXE
 * to create this file, and preserve the original copy.
 *
 * Use: CONFIG
 * with no operands to display the current configuration.
 *
 * Use: CONFIG filename
 * To install configuration values from the named file. An extension of
 * .INI is assumed if no extension is supplied.
 * Configuration element names are exactly as shown by "config" with no
 * operands, and must be followed by '=' and the new value.
 *   eg: netmask=255.255.255.0
 *       localhost=dunfield.com
 * See the example TCPIP.INI for an example.
 *
 * NOTE: This utility was written (very quickly) using my own Micro-C/PC
 *       compiler, which is a tad non-standard in a couple of areas - you
 *       should be able to easily modify it for your own tools, or simply
 *       download the free Micro-C/PC from: www.dunfield.com
 *
 * Enjoy,
 * Dave Dunfield - Sep 30, 2004
 */
#include <stdio.h>      // General I/O definitions

unsigned
    Offset = 0xA240,    // Offset into TCPIP.EXE (variable for future enhance)
    Seg,                // Segment for RAM copy
    Size,               // Size of RAM copy (Segment)
    Line;               // Current line number (for error reporting)

FILE
    *fp;                // General file pointer

unsigned char
    buffer[200],        // Line input buffer
    name[50],           // Name (from file) buffer
    word[50],           // Word (from config struct) buffer
    *ptr;               // General global pointer (for parsing)

/*
 * Structure and array describing the configuration elements
 * within TCPIP.EXE
 *
 * Note: Structure initialization is non-standard in Micro-C,
 *       essentially it is a stream of values with the keywords
 *       "char" and "int" used to switch back and forth between
 *       8-bit and 16-bit values - see compler docs for details.
 *    ** You will need to modify this for another compiler.
 */
struct Entry {
    unsigned char Type;                 // Type of element
    unsigned char *Name;                // Name of element
    unsigned Offset; } Entries[] = {    // Offset to element
    char 0, int "IP",           0x0000,
    char 0, int "NetMask",      0x0004,
    char 0, int "Gateway",      0x0008,
    char 1, int "Greenwich",    0x000C,
    char 3, int "TimeZone",     0x000E,
    char 2, int "TimeServ",     0x0012,
    char 2, int "DNS",          0x0028,
    char 3, int "Domain",       0x024E,
    char 3, int "LocalHost",    0x003E,
    char 3, int "POP3serv",     0x00CE,
    char 3, int "MailDomain",   0x010E,
    char 3, int "SMTPrelay",    0x014E,
    char 3, int "UsenetServ",   0x018E,
    char 3, int "WhoIsServ",    0x01CE,
    char 3, int "LprLpdServ",   0x020E,
    255 };  // End of list


/*
 * Display a formatted error message with input line number information
 * See Micro-C docs for use of 'register' with functions.
 */
register error(unsigned args)
{
    unsigned char buffer[100];
    _format_(nargs() + &args, buffer);
    printf("[%u]: %s\n", Line, buffer);
    exit(-1);
}

/*
 * Display a single IP address
 */
void show_ip(unsigned o)
{
    printf("%u.%u.%u.%u", peek(Seg, o), peek(Seg, ++o), peek(Seg, ++o), peek(Seg, ++o));
}

/*
 * Display a list of IP addresses (using 16 bit count value)
 */
void show_list(unsigned o)
{
    unsigned i;

    i = peek(Seg, o); o += 2;
    while(i) {
        --i;
        show_ip(o);
        o += 4;
        if(i)
            printf(" "); }
}

/*
 * Display a string from the RAM segment
 */
void show_string(unsigned o)
{
    int c;

    while(c = peek(Seg, o)) {
        putc(c, stdout);
        ++o; }
}

/*
 * Display a single entry from the configuration structure
 */
void show_entry(unsigned e)
{
    unsigned o;
    unsigned char *p;

    p = Entries[e].Name;
    for(o=0; o < 12; ++o)   // Display name in 12 char field
        putc(*p ? *p++ : ' ', stdout);

    o = Entries[e].Offset + Offset;

    switch(Entries[e].Type) {
    case 0 : show_ip(o);                    break;  // IP address
    case 1 : printf("%u", peekw(Seg, o));   break;  // 16 bit number
    case 2 : show_list(o);                  break;  // List of IPs
    case 3 : show_string(o); }                      // String

    putc('\n', stdout);
}

/*
 * Skip ahead to non-blank in input buffer
 */
int skip(void)
{
    while(isspace(*ptr))
        ++ptr;
    return *ptr;
}

/*
 * Get a decimal number from the input buffer with enforced maximum
 */
unsigned getnum(unsigned max)
{
    unsigned v, c;
    v = 0;
    skip();
    do {
        if(!isdigit(c = *ptr++))
            error("Bad number");
        v = (v * 10) + (c - '0'); }
    while(isdigit(*ptr));
    if(v > max)
        error("Out of range");
    return v;
}

/*
 * Remove expected character from input buffer - error if not found
 */
void expect(unsigned char c)
{
    if(skip() != c)
        error("expected '%c'", c);
    ++ptr;
}

/*
 * Get an IP address from the input buffer & store at offset
 */
void getip(unsigned o)
{
    poke(Seg, o, getnum(255));
    expect('.');
    poke(Seg, o+1, getnum(255));
    expect('.');
    poke(Seg, o+2, getnum(255));
    expect('.');
    poke(Seg, o+3, getnum(255));
}

/*
 * Set a single entry from the configuration structure
 */
void set_entry(unsigned e)
{
    unsigned i, j, o;
    skip();
    o = (Entries[e].Offset) + Offset;
    switch(Entries[e].Type) {
        case 0 :    getip(o);                   return; // IP
        case 1 :    pokew(Seg, o, atoi(ptr));   return; // 16bit number
        case 2 :                                        // List of IPs
            for(j=i=0; j < 22; ++j)         // Zero entire list
                poke(Seg, o + j, 0);
            j = o; o += 2;
            while(isdigit(skip())) {        // Patch in IPs
                getip(o);
                o += 4;
                ++i; }
            pokew(Seg, j, i);               // Patch in count
            return;
        case 3 :                                        // String
            i = (e == 4) ? 4 : 64;          // TimeZone is special
            for(j=0; j < i; ++j)            // Zero entire string
                poke(Seg, o + j, 0);
            while(*ptr)
                poke(Seg, o++, *ptr++); }
}       

/*
 * Main program
 */
main(int argc, char *argv[])
{
    int c;
    unsigned i;
    unsigned char *p;
    static unsigned char Changed;

    printf("CWRU-PC/IP Configuration Utility - Dave Dunfield - "#__DATE__"\n");
    // Allocate memory segment and read in TCPIP.EXE
    Seg = alloc_seg(4096);
    fp = fopen("TCPIP.EXE", "rvqb");
    while((c = getc(fp)) != EOF)
        poke(Seg, Size++, c);
    fclose(fp);

    // If no operands, display current configuration settings
    if(argc < 2) {
        for(i=0; Entries[i].Type != 255; ++i)
            show_entry(i);
        return; }

    // Open configuration file with default extension of: .INI
    strcpy(p = buffer, argv[1]);
    while(*p) {
        if(*p++ == '.')
            goto ok; }
    strcpy(p, ".INI");
ok:

    // Read config lines and set configuration values
    fp = fopen(buffer, "rvq");
    while(fgets(ptr = buffer, sizeof(buffer), fp)) {
        ++Line;
        switch(skip()) {
        case 0 :                    // Null line - ignore
        case ';' : continue; }      // Comment - ignore
        i = 0;                      // Retrieve item name in upper case
        while(isalnum(*ptr))
            name[i++] = toupper(*ptr++);
        name[i] = 0;
        expect('=');                // Skip '='
        // Extract names from structure, convert to UC and match
        skip();
        for(i=0; Entries[i].Type != 255; ++i) {
            p = Entries[i].Name;
            c = 0;
            do {
                word[c++] = toupper(*p); }
            while(*p++);
            if(!strcmp(name, word))
                goto found; }
        error("Unknown keyboard: %s\n", name);
    found:
        set_entry(i);
        Changed = 255; }
    fclose(fp);

    // If update performed, rewrite TCPIP.EXE
    if(Changed) {
        fp = fopen("TCPIP.EXE", "wvqb");
        for(i=0; i < Size; ++i)
            putc(peek(Seg, i), fp);
        fclose(fp); }
}
